#include "ds50daq/DAQ/Config.hh"
#include "ds50daq/DAQ/Aggregator.hh"
#include "art/Utilities/Exception.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "artdaq/DAQrate/EventStore.hh"
#include "art/Framework/Art/artapp.h"
#include "artdaq/DAQrate/SimpleQueueReader.hh"
#include "artdaq/Utilities/SimpleLookupPolicy.h"

#include "artdaq/DAQdata/NetMonHeader.hh"
#include "artdaq/DAQdata/RawEvent.hh"

#include <iostream>

/**
 * Constructor.
 */
ds50::Aggregator::Aggregator(int mpi_rank) :
  mpi_rank_(mpi_rank), local_group_defined_(false), art_initialized_(false),
  data_sender_count_(0), event_queue_(artdaq::getGlobalQueue())
{
  mf::LogDebug("Aggregator") << "Constructor";

  // set up an MPI communication group with other Aggregators
  int status =
    MPI_Comm_split(MPI_COMM_WORLD, ds50::Config::AggregatorTask, 0,
                   &local_group_comm_);
  if (status == MPI_SUCCESS) {
    local_group_defined_ = true;
    int temp_rank;
    MPI_Comm_rank(local_group_comm_, &temp_rank);
    mf::LogDebug("Aggregator")
      << "Successfully created local communicator for type "
      << ds50::Config::AggregatorTask << ", identifier = 0x"
      << std::hex << local_group_comm_ << std::dec
      << ", rank = " << temp_rank << ".";
  }
  else {
    mf::LogError("Aggregator")
      << "Failed to create the local MPI communicator group for "
      << "Aggregators, status code = " << status << ".";
  }
}

/**
 * Destructor.
 */
ds50::Aggregator::~Aggregator()
{
  if (local_group_defined_) {
    MPI_Comm_free(&local_group_comm_);
  }
  mf::LogDebug("Aggregator") << "Destructor";
}

/**
 * Processes the initialize request.
 */
bool ds50::Aggregator::initialize(fhicl::ParameterSet const& pset)
{
  init_string_ = pset.to_string();
  mf::LogDebug("Aggregator") << "initialize method called with DAQ "
                               << "ParameterSet = \"" << init_string_ << "\".";

  // pull out the relevant parts of the ParameterSet
  fhicl::ParameterSet daq_pset;
  try {
    daq_pset = pset.get<fhicl::ParameterSet>("daq");
  }
  catch (...) {
    mf::LogError("Aggregator")
      << "Unable to find the DAQ parameters in the initialization "
      << "ParameterSet: \"" + pset.to_string() + "\".";
    return false;
  }
  fhicl::ParameterSet agg_pset;
  try {
    agg_pset = daq_pset.get<fhicl::ParameterSet>("aggregator");
  }
  catch (...) {
    mf::LogError("Aggregator")
      << "Unable to find the aggregator parameters in the DAQ "
      << "initialization ParameterSet: \"" + daq_pset.to_string() + "\".";
    return false;
  }

  // verify that the MPI group was set up successfully
  if (! local_group_defined_) {
    mf::LogError("Aggregator")
      << "The necessary MPI group was not created in an earlier step, "
      << "and initialization can not proceed without that.";
    return false;
  }

  // determine the data receiver parameters
  try {
    max_fragment_size_words_ = daq_pset.get<uint64_t>("max_fragment_size_words");
  }
  catch (...) {
    mf::LogError("Aggregator")
      << "The max_fragment_size_words parameter was not specified "
      << "in the DAQ initialization PSet: \""
      << daq_pset.to_string() << "\".";
    return false;
  }
  try {mpi_buffer_count_ = agg_pset.get<size_t>("mpi_buffer_count");}
  catch (...) {
    mf::LogError("Aggregator")
      << "The  mpi_buffer_count parameter was not specified "
      << "in the aggregator initialization PSet: \""
      << agg_pset.to_string() << "\".";
    return false;
  }
  try {
    first_data_sender_rank_ = agg_pset.get<size_t>("first_event_builder_rank");
  }
  catch (...) {
    mf::LogError("Aggregator")
      << "The first_event_builder_rank parameter was not specified "
      << "in the aggregator initialization PSet: \""
      << agg_pset.to_string() << "\".";
    return false;
  }
  try {data_sender_count_ = agg_pset.get<size_t>("event_builder_count");}
  catch (...) {
    mf::LogError("Aggregator")
      << "The event_builder_count parameter was not specified "
      << "in the aggregator initialization PSet: \""
      << agg_pset.to_string() << "\".";
    return false;
  }
  try {
    expected_events_per_bunch_ =
      agg_pset.get<size_t>("expected_events_per_bunch");}
  catch (...) {
    mf::LogError("Aggregator")
      << "The expected_events_per_bunch parameter was not specified "
      << "in the aggregator initialization PSet: \"" << pset.to_string()
      << "\".";
    return false;
  }

  print_event_store_stats_ = agg_pset.get<bool>("print_event_store_stats", false);

  if (event_store_ptr_ == nullptr) {
    artdaq::EventStore::ART_CFGSTRING_FCN * reader = &artapp_string_config;
    event_store_ptr_.reset(new artdaq::EventStore(expected_events_per_bunch_, 1,
						  mpi_rank_, init_string_,
						  reader, print_event_store_stats_));
    event_store_ptr_->setSeqIDModulus(expected_events_per_bunch_);
    fhicl::ParameterSet tmp = pset;
    tmp.erase("daq");
    previous_pset_ = tmp;
  } else {
    fhicl::ParameterSet tmp = pset;
    tmp.erase("daq");
    if (tmp != previous_pset_) {
      mf::LogError("Aggregator")
        << "The art configuration can not be altered after art "
        << "has been configured.";
      return false;
    }
  }

  return true;
}

bool ds50::Aggregator::start(art::RunID id)
{
  run_id_ = id;
  event_store_ptr_->startRun(run_id_.run());
  return true;
}

bool ds50::Aggregator::stop()
{
  /* Nothing to do here.  The aggregator we clean up after itself once it has
     received all of the EOD fragments it expects.  Higher level code will block
     until the process_fragments() thread exits. */
  return true;
}

bool ds50::Aggregator::pause()
{
  return true;
}

bool ds50::Aggregator::resume()
{
  event_store_ptr_->startSubrun();
  return true;
}

bool ds50::Aggregator::shutdown()
{
  event_store_ptr_->endOfData();
  return true;
}

bool ds50::Aggregator::soft_initialize(fhicl::ParameterSet const& pset)
{
  mf::LogDebug("Aggregator") << "soft_initialize method called with DAQ "
                               << "ParameterSet = \"" << pset.to_string()
                               << "\".";
  return true;
}

bool ds50::Aggregator::reinitialize(fhicl::ParameterSet const& pset)
{
  mf::LogDebug("Aggregator") << "reinitialize method called with DAQ "
                               << "ParameterSet = \"" << pset.to_string()
                               << "\".";
  return true;
}

size_t ds50::Aggregator::process_fragments()
{
  size_t eodFragmentsReceived = 0;
  bool process_fragments = true;
  size_t senderSlot;
  std::vector<size_t> fragments_received(data_sender_count_ + first_data_sender_rank_, 0);
  std::vector<size_t> fragments_sent(data_sender_count_ + first_data_sender_rank_, 0);
  artdaq::FragmentPtr endSubRunMsg(nullptr);

  receiver_ptr_.reset(new artdaq::RHandles(mpi_buffer_count_,
                                           max_fragment_size_words_,
                                           data_sender_count_,
                                           first_data_sender_rank_));

  mf::LogDebug("Aggregator") << "Waiting for first fragment.";
  while (process_fragments) {
    artdaq::FragmentPtr fragmentPtr(new artdaq::Fragment);
    senderSlot = receiver_ptr_->recvFragment(*fragmentPtr);
    fragments_received[senderSlot] += 1;

    if (!art_initialized_) {
      /* The init fragment should always be the first fragment out of the
	 EventBuilder. */
      if (fragmentPtr->type() == artdaq::Fragment::InitFragmentType) {
	std::cout << "Init " << std::endl;
	artdaq::RawEvent_ptr initEvent(new artdaq::RawEvent(run_id_.run(), 1, fragmentPtr->sequenceID()));
	initEvent->insertFragment(std::move(fragmentPtr));
	event_queue_.enqNowait(initEvent);
	art_initialized_ = true;
      }
    } else {
      /* Note that in the currently implementation of the NetMon output/input
	 modules there are no EndOfRun or Shutdown fragments. */
      if (fragmentPtr->type() == artdaq::Fragment::DataFragmentType) {
	event_store_ptr_->insert(std::move(fragmentPtr));
      } else if (fragmentPtr->type() == artdaq::Fragment::EndOfSubrunFragmentType) {
	/* We inject the EndSubrun fragment after all other data has been
	   received.  The SHandles and RHandles classes do not guarantee that 
	   data will be received in the same order it is sent.  We'll hold on to
	   this fragment and inject it once we've received all EOD fragments. */
	endSubRunMsg = std::move(fragmentPtr);
      } else if (fragmentPtr->type() == artdaq::Fragment::EndOfDataFragmentType) {
	eodFragmentsReceived++;
      /* We count the EOD fragment as a fragment received but the SHandles class
	 does not count it as a fragment sent which means we need to add one to
	 the total expected fragments. */
	fragments_sent[senderSlot] = *fragmentPtr->dataBegin() + 1;
      }
    }

    /* If we've received EOD fragments from all of the EventBuilders we can
       verify that we've also received every fragment that they have sent.  If
       all fragments are accounted for we can flush the EventStoreand exit out 
       of this thread.*/
    if (eodFragmentsReceived == data_sender_count_ && endSubRunMsg != nullptr) {
      bool fragmentsOutstanding = false;
      for (size_t i = 0; i < data_sender_count_ + first_data_sender_rank_; i++) {
	if (fragments_received[i] != fragments_sent[i]) {
          fragmentsOutstanding = true;
          break;
        }
      }

      if (!fragmentsOutstanding) {
	event_store_ptr_->flushData();
	artdaq::RawEvent_ptr subRunEvent(new artdaq::RawEvent(run_id_.run(), 1, 0));
	subRunEvent->insertFragment(std::move(endSubRunMsg));
	event_queue_.enqNowait(subRunEvent);
	process_fragments = false;
      }
    }
  }

  receiver_ptr_.reset(nullptr);
  return 0;
}

std::string ds50::Aggregator::report(std::string const&) const
{
  // lots of cool stuff that we can do here
  // - report on the number of fragments received and the number
  //   of events built (in the current or previous run
  // - report on the number of incomplete events in the EventStore
  //   (if running)
  std::string tmpString = "Event Builder run number = ";
  tmpString.append(boost::lexical_cast<std::string>(run_id_.run()));
  return tmpString;
}
