#!/usr/bin/env ruby

require "open3"

def fileInUse(filePath)
  # Determine if a file is in use using the lsof utility.  If the file isn't in
  # use the utility will return no output to stdout.
  output = nil
  Open3.popen3("lsof %s" % [filePath]) { |stdin, stdout, stderr|
    output = stdout.read
  }

  if output == ""
      return false
  end

  return true
end

def getMetaData(filePath)
  # Run the dataMetaData utility over the given file.  If the daqMetaData utility
  # doesn't return anything (either the file is empty or corrupt) return nil to
  # the caller.  If actual meta data is returned format it into a hash and
  # return it.
  output = nil
  Open3.popen3("daqMetaData %s" % [filePath]) { |stdin, stdout, stderr|
    output = stdout.read
  }

  lines = output.split("\n")
  if lines.length == 1
    return nil
  end

  firstEvent = lines[0].split("\t")
  lastEvent = lines[1].split("\t")

  return Hash["firstRun" => firstEvent[0], "lastRun" => lastEvent[0],
              "firstSubRun" => firstEvent[1], "lastSubRun" => lastEvent[1],
              "firstEvent" => firstEvent[2], "lastEvent" => lastEvent[2]]
end

def getUUID()
  # There is now UUID module in the version of ruby that we are using so we'll
  # use the command line utility.
  output = nil
  Open3.popen3("uuidgen") { |stdin, stdout, stderr|
    output = stdout.read
  }

  return output.chomp
end

if __FILE__ == $0
  if ARGV.length != 1
    puts "Usage:"
    puts "  %s /path/to/data/directory" % [__FILE__]
    exit
  end

  fileEntries = Dir.entries(ARGV[0])
  fileEntries.each do |fileEntry|
    # Assume that the DAQ writes out files starting with "ds50daq".
    if fileEntry =~ /^ds50daq/
      fullPath = File.join(ARGV[0], fileEntry)
      if fileInUse(fullPath)
        puts "Skipping %s, in use." % [fileEntry]
        next
      end

      metaData = getMetaData(fullPath)
      if metaData == nil
        puts "Deleting %s" % [fileEntry]
        File.delete(fullPath)
      else
        # I think it's a good idea to include a UUID in each file as I can think
        # of valid use cases where we'd want to duplicate events to different
        # files and run-subrun-event wouldn't be a unique identifier.
        newName = "ds50data_run%s_subrun%s_event%s-run%s_subrun%s_event%s-%s.root" % [metaData["firstRun"],
                                                                                      metaData["firstSubRun"],
                                                                                      metaData["firstEvent"],
                                                                                      metaData["lastRun"],
                                                                                      metaData["lastSubRun"],
                                                                                      metaData["lastEvent"],
                                                                                      getUUID()]
        puts "Renaming %s to %s" % [fileEntry, newName]
        File.rename(fullPath, File.join(ARGV[0], newName))
      end
    end
  end
end
